/**
   Copyright 2010 Carlo Alberto Degli Atti (lord kada)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package it.concept.pin.proxy;

import it.concept.pin.PinManager;
import it.concept.pin.annotations.Model;
import it.concept.pin.annotations.ProxyAdapter;

import java.lang.reflect.*;
import java.util.Collection;

public class ProxyWrapper implements InvocationHandler {

	private Object target;

	public static Object createProxy( Class<?> interfaceName, Object target ) throws InstantiationException, IllegalAccessException, SecurityException, NoSuchMethodException, IllegalArgumentException, InvocationTargetException {

		Object targetObject = target;
		
		ProxyAdapter proxyAnnotation = targetObject.getClass().getAnnotation( ProxyAdapter.class );
		
		if ( proxyAnnotation != null ) {
			
			Constructor<?> proxyConstructor = proxyAnnotation.value().getConstructor( new Class<?>[] { targetObject.getClass() } );
		
			targetObject = proxyConstructor.newInstance( targetObject );
			
		}
			
		return Proxy.newProxyInstance( interfaceName.getClassLoader(), new Class[] { interfaceName }, new ProxyWrapper( targetObject ) );

	}

	/**
	 * Provate constructor
	 * @param target
	 */
	private ProxyWrapper( Object target ) {

		this.target = target;

	}
	
	@SuppressWarnings("unchecked")
	public Object invoke( Object proxy, Method method, Object[] args ) {

		Object result = null;

		Class<?> argClasses[] = null;

		if ( args != null ) {

			argClasses = new Class<?>[ args.length ];

			for ( int i = 0; i < args.length; i++ )
				argClasses[ i ] = args[ i ].getClass();

		}

		try {

			try {
				
				Method proxyMethod = target.getClass().getMethod( method.getName(), argClasses );
				
				result = proxyMethod.invoke( target, args );
			
			} catch ( NoSuchMethodException e ) {
				
				if ( target instanceof AbstractProxyAdapter<?> ) {

					Object pojo = ( ( AbstractProxyAdapter<?> ) target ).getPojo();
					Method pojoMethod = pojo.getClass().getMethod( method.getName(), argClasses );
					result = pojoMethod.invoke( ( ( AbstractProxyAdapter<?> ) target ).getPojo(), args );

				}
				
				else
					
					throw e;
					
			}
				
			if ( result != null ) {
				
				Model pinAnnotation = result.getClass().getAnnotation( Model.class );
				
				if ( pinAnnotation != null ) 
					
					return PinManager.toModel( result );
					
				else
					
					if ( result instanceof Collection<?> ) {
						
						Collection<?> resultCollection = ( Collection<?> ) result;
						
						if ( resultCollection.size() > 0 ) {

							Model resultPinAnnotation = resultCollection.toArray()[0].getClass().getAnnotation( Model.class );
							
							if ( resultPinAnnotation != null ) {
								
								Collection<Object> newInstance = ( Collection<Object> ) result.getClass().newInstance();
								
								for ( Object item : resultCollection )
									
									newInstance.add( PinManager.toModel( item ) );
								
								return newInstance;
								
							}

						}
						
					}
				
			}
			
			return result;

		} catch ( Exception e ) {

			throw new RuntimeException( e );
			
		}

	}

}